/* 
The NIST RCS (Real-time Control Systems) 
 library is public domain software, however it is preferred
 that the following disclaimers be attached.

Software Copywrite/Warranty Disclaimer

   This software was developed at the National Institute of Standards and
Technology by employees of the Federal Government in the course of their
official duties. Pursuant to title 17 Section 105 of the United States
Code this software is not subject to copyright protection and is in the
public domain. NIST Real-Time Control System software is an experimental
system. NIST assumes no responsibility whatsoever for its use by other
parties, and makes no guarantees, expressed or implied, about its
quality, reliability, or any other characteristic. We would appreciate
acknowledgement if the software is used. This software can be
redistributed and/or modified freely provided that any derivative works
bear some notice that they are derived from it, and any modified
versions bear some notice that they have been modified.



*/ 

/*
   _posemath.c

   C definitions for pose math library data types and
   manipulation functions.

   Modification history:

   15-Jan-1998  FMP added pmHomInv()
   5-Jan-1998  FMP added pmAxisAngleQuatConvert(), pmQuatAxisAngleMult()
   10-Oct-1997  FMP chickened out and checked for divide-by-0 always,
   not just if PM_DEBUG is defined. Set results to DBL_MAX where appropriate.
   17-Aug-1997  FMP removed include mathprnt.h-- not needed here
   11-Jul-1997  FMP changed names from PM_CARTESIAN, for example, to
   PmCartesian, to get rid of name conflicts with mixed C/C++ apps
   19-Jun-1997  FMP added pmLine, pmCircle stuff
   18-Jun-1997  FMP added pmCartPlaneProj()
   14-Apr-1997  FMP created from C parts of posemath.c
*/

package rcs.posemath;

public class Posemath
{

  /* global error number */
  static public int pmErrno = 0;
  static public boolean PM_PRINT_ERROR = false;
  static public boolean debug_on = false;

  // Miscellaneous Pre-Defined Values
  public static final double CART_FUZZ  = 0.0001;
  public static final double DOUBLECP_FUZZ  = 1.0842021724855044e-19;
  public static final double DOUBLE_FUZZ  = 2.2204460492503131e-16;
  public static final double E_EPSILON  = .005;
  public static final double PM_2_PI  = 6.28318530717958647692;
  public static final int PM_DIV_ERR  = -4;
  public static final int PM_ERR  = -1;
  public static final int PM_IMPL_ERR  = -2;
  public static final int PM_NORM_ERR  = -3;
  public static final double PM_PI  = 3.14159265358979323846;
  public static final double PM_PI_2  = 1.57079632679489661923;
  public static final double PM_PI_4  = 0.78539816339744830962;
  public static final double QSIN_FUZZ  = .0003;
  public static final double QS_FUZZ  = .0003;
  public static final double Q_FUZZ  = .0003;
  public static final double RPY_P_FUZZ  = 0.001;
  public static final double RS_FUZZ  = .0001;
  public static final double SINGULAR_EPSILON  = .0001;
  public static final double SQRT_FUZZ  = -.0002;
  public static final double UNIT_QUAT_FUZZ  = .01;
  public static final double UNIT_SC_FUZZ  = .01;
  public static final double UNIT_VEC_FUZZ  = .01;
  public static final double V_FUZZ  = .0005;
  public static final double ZYX_Y_FUZZ  = 0.001;
  public static final double ZYZ_Y_FUZZ  = 0.001;
  public static final double DBL_MAX = 2^969;


  public static final double PI = PM_PI;
  public static final double PI_2 = PM_PI_2;
  public static final double PI_4 = PM_PI_4;

  // Enumerated Type Constants

  // PmAxis
  public static final int PM_X=0;
  public static final int PM_Y=1;
  public static final int PM_Z=2;


  static protected void pmPrintError(String str, int new_errno) throws PmException
  {
    pmErrno = new_errno;
    if(PM_PRINT_ERROR)
      {
        try
          {
            System.err.println(str);
          }
        catch(Exception e)
          {
            e.printStackTrace();
          }
      }
    throw new PmException(pmErrno, str);
  }

  /* fuzz checker */
  static public boolean IS_FUZZ(double a, double fuzz) throws PmException
  {
    return Math.abs(a) < fuzz;
  }


  /* Pose Math Basis Functions */

  /* Scalar functions */

  static public double pmSqrt(double x) throws PmException
  {
    if (x > 0.0)
      {
        return Math.sqrt(x);
      }

    if (x > SQRT_FUZZ)
      {
        return 0.0;
      }

    pmPrintError("sqrt of large negative number\n", PM_ERR);

    return 0.0;
  }

  static public double pmSq(double x) throws PmException
  {
    return x*x;
  }

  public static boolean pmClose(double a, double b, double eps)
  {
    return(Math.abs(a - b) < eps);
  }


  /* Translation rep conversion functions */

  static public PM_CYLINDRICAL toCyl(PM_CARTESIAN v) throws PmException
  {
    PM_CYLINDRICAL c = new PM_CYLINDRICAL();
    pmCartCylConvert(v, c);
    return c;
  }

  static public int pmCartCylConvert(PmCartesian v, PmCylindrical c) throws PmException
  {
    c.theta = Math.atan2(v.y, v.x);
    c.r = pmSqrt(pmSq(v.x) + pmSq(v.y));
    c.z = v.z;

    return pmErrno = 0;
  }

  static public PM_SPHERICAL toSph(PM_CARTESIAN v) throws PmException
  {
    PM_SPHERICAL s = new PM_SPHERICAL();
    pmCartSphConvert(v, s);
    return s;
  }

  static public int pmCartSphConvert(PmCartesian v, PmSpherical s) throws PmException
  {
    double _r;

    s.theta = Math.atan2(v.y, v.x);
    s.r = pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z));
    _r = pmSqrt(pmSq(v.x) + pmSq(v.y));
    s.phi = Math.atan2(_r, v.z);

    return pmErrno = 0;
  }

  static public PM_CARTESIAN toCart(PM_SPHERICAL v) throws PmException
  {
    PM_CARTESIAN c = new PM_CARTESIAN();
    pmSphCartConvert(v,c);
    return c;
  }

  static public int pmSphCartConvert(PmSpherical s, PmCartesian  v) throws PmException
  {
    double _r;

    _r = s.r * Math.sin(s.phi);
    v.z = s.r * Math.cos(s.phi);
    v.x = _r * Math.cos(s.theta);
    v.y = _r * Math.sin(s.theta);

    return pmErrno = 0;
  }

  static public PM_CYLINDRICAL toCyl(PM_SPHERICAL s) throws PmException
  {
    PM_CYLINDRICAL  c = new PM_CYLINDRICAL();
    pmSphCylConvert(s, c);
    return c;
  }

  static public int pmSphCylConvert(PmSpherical s, PmCylindrical  c) throws PmException
  {
    PmCartesian v = new PmCartesian();
    pmSphCartConvert(s,v);
    pmCartCylConvert(v,c);
    return pmErrno;
  }

  static public PM_CARTESIAN toCart(PM_CYLINDRICAL v) throws PmException
  {
    PM_CARTESIAN c = new PM_CARTESIAN();
    pmCylCartConvert(v, c);
    return c;
  }

  static public int pmCylCartConvert(PmCylindrical c, PmCartesian v) throws PmException
  {
    v.x = c.r * Math.cos(c.theta);
    v.y = c.r * Math.sin(c.theta);
    v.z = c.z;

    return pmErrno = 0;
  }

  static public PM_SPHERICAL toSph(PM_CYLINDRICAL c) throws PmException
  {
    PM_SPHERICAL s = new PM_SPHERICAL();
    pmCylSphConvert(c, s);
    return s;
  }

  static public int pmCylSphConvert(PmCylindrical c, PmSpherical s) throws PmException
  {
    pmErrno = 0;
    PmCartesian v = new PmCartesian();
    pmCylCartConvert(c,v);
    pmCartSphConvert(v,s);
    return pmErrno;
  }


  /* Rotation rep conversion functions */
  static public int pmAxisAngleQuatConvert(int axis, double a, PmQuaternion q) throws PmException
  {
    double sh;

    a *= 0.5;
    sh = Math.sin(a);
    q.s = Math.cos(a);

    switch (axis)
      {
      case PM_X:
        q.x = sh;
        q.y = 0.0;
        q.z = 0.0;
        break;

      case PM_Y:
        q.x = 0.0;
        q.y = sh;
        q.z = 0.0;
        break;

      case PM_Z:
        q.x = 0.0;
        q.y = 0.0;
        q.z = sh;
        break;

      default:
        pmPrintError("error: bad axis in pmAxisAngleQuatConvert\n", PM_ERR);
        return -1;
      }

    if( q.s < 0.0 )
      {
        q.s *= -1.0;
        q.x *= -1.0;
        q.y *= -1.0;
        q.z *= -1.0;
      }

    return 0;
  }


  static public PM_QUATERNION toQuat(PM_ROTATION_VECTOR v) throws PmException
  {
    PM_QUATERNION q = new PM_QUATERNION();
    pmRotQuatConvert(v,q);
    return q;
  }

  static public int pmRotQuatConvert(PmRotationVector r, PmQuaternion  q) throws PmException
  {
    double sh;

    /* make sure r is normalized */
    if (0 != pmRotNorm(r, r))
      {

        pmPrintError("error: pmRotQuatConvert rotation vector not normalized\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    if (pmClose(r.s, 0.0, 1e-3))
      {
        q.s = 1.0;
        q.x = q.y = q.z = 0.0;

        return pmErrno = 0;
      }

    sh = Math.sin(r.s/2.0);
    q.s = Math.cos(r.s/2.0);

    if (q.s >= 0.0)
      {
        q.x = r.x * sh;
        q.y = r.y * sh;
        q.z = r.z * sh;
      }
    else
      {
        q.s *= -1;
        q.x = -r.x * sh;
        q.y = -r.y * sh;
        q.z = -r.z * sh;
      }

    return pmErrno = 0;
  }


  static public PM_ROTATION_MATRIX toMat(PM_ROTATION_VECTOR v) throws PmException
  {
    PM_ROTATION_MATRIX m = new PM_ROTATION_MATRIX();
    pmRotMatConvert(v,m);
    return m;
  }

 static public int pmRotMatConvert(PmRotationVector r, PmRotationMatrix m) throws PmException
  {
    double s, c, omc;


    if (! pmRotIsNorm(r))
      {

        pmPrintError("Bad vector in pmRotMatConvert\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }

    s = Math.sin(r.s);
    c = Math.cos(r.s);

    /* from space book */
    m.x.x = c + pmSq(r.x) * (omc = 1 - c);      /* omc = One Minus Cos */
    m.y.x = -r.z * s + r.x * r.y * omc;
    m.z.x = r.y * s + r.x * r.z * omc;

    m.x.y = r.z * s + r.y * r.x * omc;
    m.y.y = c + pmSq(r.y) * omc;
    m.z.y = -r.x * s + r.y * r.z * omc;

    m.x.z = -r.y * s + r.z * r.x * omc;
    m.y.z = r.x * s + r.z * r.y * omc;
    m.z.z = c + pmSq(r.z) * omc;

    return pmErrno = 0;
  }

  static public int pmRotZyzConvert(PmRotationVector r, PmEulerZyz zyz) throws PmException
  {

    pmPrintError("error: pmRotZyzConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmRotZyxConvert(PmRotationVector r, PmEulerZyx zyx) throws PmException
  {

    pmPrintError("error: pmRotZyxConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmRotRpyConvert(PmRotationVector r, PmRpy rpy) throws PmException
  {

    pmPrintError("error: pmRotRpyConvert not implemented\n", PM_IMPL_ERR );

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_ROTATION_VECTOR toRot(PM_QUATERNION v) throws PmException
  {
    PM_ROTATION_VECTOR r = new PM_ROTATION_VECTOR();
    pmQuatRotConvert(v,r);
    return r;
  }


  static public int pmQuatRotConvert(PmQuaternion q, PmRotationVector r) throws PmException
  {
    double sh;


    if (! pmQuatIsNorm(q))
      {

        pmPrintError("Bad quaternion in pmQuatRotConvert\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    sh = pmSqrt(pmSq(q.x) + pmSq(q.y) + pmSq(q.z));

    if (sh > QSIN_FUZZ)
      {
        r.s = 2.0 * Math.atan2(sh, q.s);
        r.x = q.x / sh;
        r.y = q.y / sh;
        r.z = q.z / sh;
      }
    else
      {
        r.s = 0.0;
        r.x = 0.0;
        r.y = 0.0;
        r.z = 0.0;
      }

    return pmErrno = 0;
  }

  static public PM_ROTATION_MATRIX toMat(PM_QUATERNION v) throws PmException
  {
    PM_ROTATION_MATRIX m = new PM_ROTATION_MATRIX();
    pmQuatMatConvert(v,m);
    return m;
  }

  static public int pmQuatMatConvert(PmQuaternion q, PmRotationMatrix m) throws PmException
  {

    if (! pmQuatIsNorm(q))
      {

        pmPrintError("Bad quaternion in pmQuatMatConvert\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    /* from space book where e1=q.x e2=q.y e3=q.z e4=q.s */
    m.x.x = 1 - 2 * (pmSq(q.y) + pmSq(q.z));
    m.y.x = 2 * (q.x * q.y - q.z * q.s);
    m.z.x = 2 * (q.z * q.x + q.y * q.s);

    m.x.y = 2 * (q.x * q.y + q.z * q.s);
    m.y.y = 1 - 2 * (pmSq(q.z) + pmSq(q.x));
    m.z.y = 2 * (q.y * q.z - q.x * q.s);

    m.x.z = 2 * (q.z * q.x - q.y * q.s);
    m.y.z = 2 * (q.y * q.z + q.x * q.s);
    m.z.z = 1 - 2 * (pmSq(q.x) + pmSq(q.y));

    return pmErrno = 0;
  }

  static public int pmQuatZyzConvert(PmQuaternion q, PmEulerZyz zyz) throws PmException
  {
    PmRotationMatrix m  = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmQuatMatConvert(q, m);
    r2 = pmMatZyzConvert(m, zyz);

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmQuatZyxConvert(PmQuaternion q, PmEulerZyx zyx) throws PmException
  {
    PmRotationMatrix m = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmQuatMatConvert(q, m);
    r2 = pmMatZyxConvert(m, zyx);

     if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;

  }

  static public PM_RPY toRpy(PM_QUATERNION v) throws PmException
  {
    PM_RPY r = new PM_RPY();
    pmQuatRpyConvert(v,r);
    return r;
  }

  static public int pmQuatRpyConvert(PmQuaternion q, PmRpy rpy) throws PmException
  {
    PmRotationMatrix m = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmQuatMatConvert(q, m);
    r2 = pmMatRpyConvert(m, rpy);

     if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmMatRotConvert(PmRotationMatrix m, PmRotationVector r) throws PmException
  {

    pmPrintError("error: pmMatRotConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_QUATERNION toQuat(PM_ROTATION_MATRIX m) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmMatQuatConvert(m,qout);
    return qout;
  }

  static public int pmMatQuatConvert(PmRotationMatrix m, PmQuaternion q) throws PmException
  {
    /*
      from Stephe's "space" book
      e1 = (c32 - c23) / 4*e4
      e2 = (c13 - c31) / 4*e4
      e3 = (c21 - c12) / 4*e4
      e4 = sqrt(1 + c11 + c22 + c33) / 2

      if e4 == 0
      e1 = sqrt(1 + c11 - c33 - c22) / 2
      e2 = sqrt(1 + c22 - c33 - c11) / 2
      e3 = sqrt(1 + c33 - c11 - c22) / 2
      to determine whether to take the positive or negative sqrt value
      since e4 == 0 indicates a 180* rotation then (0 x y z) = (0 -x -y -z).
      Thus some generallities can be used:
      1) find which of e1, e2, or e3 has the largest magnitude and leave it pos.
      2) if e1 is largest then
      if c21 < 0 then take the negative for e2
      if c31 < 0 then take the negative for e3
      3) else if e2 is largest then
      if c21 < 0 then take the negative for e1
      if c32 < 0 then take the negative for e3
      4) else if e3 is larget then
      if c31 < 0 then take the negative for e1
      if c32 < 0 then take the negative for e2

      Note: c21 in the space book is m.x.y in this C code
      */

    double a;


    if (! pmMatIsNorm(m))
      {

        pmPrintError("Bad matrix in pmMatQuatConvert\n",PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    q.s = 0.5 * pmSqrt(1.0 + m.x.x + m.y.y + m.z.z);

    if (Math.abs(q.s) > QS_FUZZ)
      {
        q.x = (m.y.z - m.z.y) / (a = 4 * q.s);
        q.y = (m.z.x - m.x.z) / a;
        q.z = (m.x.y - m.y.x) / a;
      }
    else
      {
        q.s = 0;
        q.x = pmSqrt(1.0 + m.x.x - m.y.y - m.z.z) / 2.0;
        q.y = pmSqrt(1.0 + m.y.y - m.x.x - m.z.z) / 2.0;
        q.z = pmSqrt(1.0 + m.z.z - m.y.y - m.x.x) / 2.0;

        if (q.x > q.y &&  q.x > q.z)
          {
            if (m.x.y < 0.0)
              {
                q.y *= -1;
              }
            if (m.x.z < 0.0)
              {
                q.z *= -1;
              }
          }
        else if (q.y > q.z)
          {
            if (m.x.y < 0.0)
              {
                q.x *= -1;
              }
            if (m.y.z < 0.0)
              {
                q.z *= -1;
              }
          }
        else
          {
            if (m.x.z < 0.0)
              {
                q.x *= -1;
              }
            if (m.y.z < 0.0)
              {
                q.y *= -1;
              }
          }
      }

    return pmQuatNorm(q, q);
  }

  static public int pmMatZyzConvert(PmRotationMatrix m, PmEulerZyz zyz) throws PmException
  {
    zyz.y = Math.atan2(pmSqrt(pmSq(m.x.z) + pmSq(m.y.z)), m.z.z);

    if (Math.abs(zyz.y) < ZYZ_Y_FUZZ)
      {
        zyz.z = 0.0;
        zyz.y = 0.0;            /* force Y to 0 */
        zyz.zp = Math.atan2(-m.y.x, m.x.x);
      }
    else if (Math.abs(zyz.y - PM_PI) < ZYZ_Y_FUZZ)
      {
        zyz.z = 0.0;
        zyz.y = PM_PI;          /* force Y to 180 */
        zyz.zp = Math.atan2(m.y.x, -m.x.x);
      }
    else
      {
        zyz.z = Math.atan2(m.z.y, m.z.x);
        zyz.zp = Math.atan2(m.y.z, -m.x.z);
      }

    return pmErrno = 0;
  }

  static public int pmMatZyxConvert(PmRotationMatrix m, PmEulerZyx zyx) throws PmException
  {
    zyx.y = Math.atan2(-m.x.z, pmSqrt(pmSq(m.x.x) + pmSq(m.x.y)));

    if (Math.abs(zyx.y - PM_PI_2) < ZYX_Y_FUZZ)
      {
        zyx.z = 0.0;
        zyx.y = PM_PI_2;                /* force it */
        zyx.x = Math.atan2(m.y.x, m.y.y);
      }
    else if (Math.abs(zyx.y + PM_PI_2) < ZYX_Y_FUZZ)
      {
        zyx.z = 0.0;
        zyx.y = -PM_PI_2;               /* force it */
        zyx.x = -Math.atan2(m.y.z, m.y.y);
      }
    else
      {
        zyx.z = Math.atan2(m.x.y, m.x.x);
        zyx.x = Math.atan2(m.y.z, m.z.z);
      }

    return pmErrno = 0;
  }

  static public int pmMatRpyConvert(PmRotationMatrix m, PmRpy rpy) throws PmException
  {
    rpy.p = Math.atan2(-m.x.z, pmSqrt(pmSq(m.x.x) + pmSq(m.x.y)));

    if (Math.abs(rpy.p -PM_PI_2) < RPY_P_FUZZ)
      {
        rpy.r = Math.atan2(m.y.x, m.y.y);
        rpy.p =PM_PI_2;         /* force it */
        rpy.y = 0.0;
      }
    else if (Math.abs(rpy.p +PM_PI_2) < RPY_P_FUZZ)
      {
        rpy.r = -Math.atan2(m.y.z, m.y.y);
        rpy.p = -PM_PI_2;               /* force it */
        rpy.y = 0.0;
      }
    else
      {
        rpy.r = Math.atan2(m.y.z, m.z.z);
        rpy.y = Math.atan2(m.x.y, m.x.x);
      }

    return pmErrno = 0;
  }

  static public int pmZyzRotConvert(PmEulerZyz zyz, PmRotationVector r) throws PmException
  {

    pmPrintError("error: pmZyzRotConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_QUATERNION toQuat(PM_EULER_ZYZ v) throws PmException
  {
    PM_QUATERNION q = new PM_QUATERNION();
    pmZyzQuatConvert(v,q);
    return q;
  }

  static public int pmZyzQuatConvert(PmEulerZyz zyz, PmQuaternion q) throws PmException
  {
    PmRotationMatrix m = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmZyzMatConvert(zyz, m);
    r2 = pmMatQuatConvert(m, q);

      if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmZyzMatConvert(PmEulerZyz zyz, PmRotationMatrix m) throws PmException
  {
    double sa, sb, sg;
    double ca, cb, cg;

    sa = Math.sin(zyz.z);
    sb = Math.sin(zyz.y);
    sg = Math.sin(zyz.zp);

    ca = Math.cos(zyz.z);
    cb = Math.cos(zyz.y);
    cg = Math.cos(zyz.zp);

    m.x.x = ca * cb * cg - sa * sg;
    m.y.x = -ca * cb * sg - sa * cg;
    m.z.x = ca * sb;

    m.x.y = sa * cb * cg + ca * sg;
    m.y.y = -sa * cb * sg + ca * cg;
    m.z.y = sa * sb;

    m.x.z = -sb * cg;
    m.y.z = sb * sg;
    m.z.z = cb;

    return pmErrno = 0;
  }

  static public int pmZyzRpyConvert(PmEulerZyz zyz, PmRpy rpy) throws PmException
  {

    pmPrintError("error: pmZyzRpyConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmZyxRotConvert(PmEulerZyx zyx, PmRotationVector r) throws PmException
  {

    pmPrintError("error: pmZyxRotConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_QUATERNION toQuat(PM_EULER_ZYX v) throws PmException
  {
    PM_QUATERNION q = new PM_QUATERNION();
    pmZyxQuatConvert(v,q);
    return q;
  }


  static public int pmZyxQuatConvert(PmEulerZyx zyx, PmQuaternion q) throws PmException
  {
    PmRotationMatrix m = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmZyxMatConvert(zyx, m);
    r2 = pmMatQuatConvert(m, q);

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmZyxMatConvert(PmEulerZyx zyx, PmRotationMatrix m) throws PmException
  {
    double sa, sb, sg;
    double ca, cb, cg;

    sa = Math.sin(zyx.z);
    sb = Math.sin(zyx.y);
    sg = Math.sin(zyx.x);

    ca = Math.cos(zyx.z);
    cb = Math.cos(zyx.y);
    cg = Math.cos(zyx.x);

    m.x.x = ca * cb;
    m.y.x = ca * sb * sg - sa * cg;
    m.z.x = ca * sb * cg + sa * sg;

    m.x.y = sa * cb;
    m.y.y = sa * sb * sg + ca * cg;
    m.z.y = sa * sb * cg - ca * sg;

    m.x.z = -sb;
    m.y.z = cb * sg;
    m.z.z = cb * cg;

    return pmErrno = 0;
  }

  static public int pmZyxZyzConvert(PmEulerZyx zyx, PmEulerZyz zyz) throws PmException
  {

    pmPrintError("error: pmZyxZyzConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmZyxRpyConvert(PmEulerZyx zyx, PmRpy rpy) throws PmException
  {

    pmPrintError("error: pmZyxRpyConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmRpyRotConvert(PmRpy rpy, PmRotationVector r) throws PmException
  {

    pmPrintError("error: pmRpyRotConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_QUATERNION toQuat(PM_RPY v) throws PmException
  {
    PM_QUATERNION q = new PM_QUATERNION();
    pmRpyQuatConvert(v,q);
    return q;
  }

  static public int pmRpyQuatConvert(PmRpy rpy, PmQuaternion q) throws PmException
  {
    PmRotationMatrix m = new PmRotationMatrix();
    int r1, r2;

    /* FIXME-- need direct equations */
    r1 = pmRpyMatConvert(rpy, m);
    r2 = pmMatQuatConvert(m, q);

     if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmRpyMatConvert(PmRpy rpy, PmRotationMatrix m) throws PmException
  {
    double sa, sb, sg;
    double ca, cb, cg;

    sa = Math.sin(rpy.y);
    sb = Math.sin(rpy.p);
    sg = Math.sin(rpy.r);

    ca = Math.cos(rpy.y);
    cb = Math.cos(rpy.p);
    cg = Math.cos(rpy.r);

    m.x.x = ca * cb;
    m.y.x = ca * sb * sg - sa * cg;
    m.z.x = ca * sb * cg + sa * sg;

    m.x.y = sa * cb;
    m.y.y = sa * sb * sg + ca * cg;
    m.z.y = sa * sb * cg - ca * sg;

    m.x.z = -sb;
    m.y.z = cb * sg;
    m.z.z = cb * cg;

    return pmErrno = 0;
  }

  static public int pmRpyZyzConvert(PmRpy rpy, PmEulerZyz zyz) throws PmException
  {

    pmPrintError("error: pmRpyZyzConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public int pmRpyZyxConvert(PmRpy rpy, PmEulerZyx zyx) throws PmException
  {

    pmPrintError("error: pmRpyZyxConvert not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public PM_HOMOGENEOUS toHom(PM_POSE p) throws PmException
  {
    PM_HOMOGENEOUS hout = new PM_HOMOGENEOUS();
    pmPoseHomConvert(p,hout);
    return hout;
  }

  static public int pmPoseHomConvert(PmPose p, PmHomogeneous h) throws PmException
  {
    int r1;

    h.tran = p.tran;
    r1 = pmQuatMatConvert(p.rot, h.rot);

    return pmErrno = r1;
  }

  static public int pmHomPoseConvert(PmHomogeneous h, PmPose p) throws PmException
  {
    int r1;

    p.tran = h.tran;
    r1 = pmMatQuatConvert(h.rot, p.rot);

    return pmErrno = r1;
  }

  /* PmCartesian functions */
  static public boolean pmCartCartCompare(PmCartesian v1, PmCartesian v2) throws PmException
  {
    if(debug_on)
      {
        System.out.println("pmCartCartCompare("+v1+", "+v2+")");
      }
    if (Math.abs(v1.x-v2.x) >= V_FUZZ ||
        Math.abs(v1.y-v2.y) >= V_FUZZ ||
        Math.abs(v1.z-v2.z) >= V_FUZZ)
      {
        return false;
      }

    return true;
  }

  static public double dot(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    return pmCartCartDot(v1,v2);
  }

  static public double pmCartCartDot(PmCartesian v1, PmCartesian v2) throws PmException
  {
    double d = v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;

    return d;
  }


  static public PM_CARTESIAN cross(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartCartCross(v1,v2, vout);
    return vout;
  }

  static public int  pmCartCartCross(PmCartesian v1, PmCartesian v2, PmCartesian vout) throws PmException
  {
    vout.x = v1.y * v2.z - v1.z * v2.y;
    vout.y = v1.z * v2.x - v1.x * v2.z;
    vout.z = v1.x * v2.y - v1.y * v2.x;

    pmErrno = 0;
    return 0;
  }

  static public double mag(PM_CARTESIAN v) throws PmException
  {
    return pmCartMag(v);
  }

  static public double  pmCartMag(PmCartesian v) throws PmException
  {
    double d = pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z));
    return d;
  }

  static public double disp(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    return pmCartCartDisp(v1,v2);
  }

  static public double pmCartCartDisp(PmCartesian v1, PmCartesian v2) throws PmException
  {
    double d = pmSqrt(pmSq(v2.x - v1.x) + pmSq(v2.y - v1.y) + pmSq(v2.z - v1.z));
    return d;
  }

  static public PM_CARTESIAN add(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartCartAdd(v1,v2, vout);
    return vout;
  }

  static public int  pmCartCartAdd(PmCartesian v1, PmCartesian v2, PmCartesian vout ) throws PmException
  {
    vout.x = v1.x + v2.x;
    vout.y = v1.y + v2.y;
    vout.z = v1.z + v2.z;

     pmErrno = 0;
     return 0;
  }

  static public PM_CARTESIAN subtract(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartCartSub(v1,v2, vout);
    return vout;
  }

  static public int pmCartCartSub(PmCartesian v1, PmCartesian v2, PmCartesian vout) throws PmException
  {
    vout.x = v1.x - v2.x;
    vout.y = v1.y - v2.y;
    vout.z = v1.z - v2.z;

    pmErrno = 0;
    return 0;
  }

  static public PM_CARTESIAN multiply(double d, PM_CARTESIAN v) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartScalMult(v,d, vout);
    return vout;
  }

  static public PM_CARTESIAN multiply(PM_CARTESIAN v, double d) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartScalMult(v,d, vout);
    return vout;
  }


  static public int pmCartScalMult(PmCartesian v1, double d, PmCartesian vout) throws PmException
  {
    vout.x = v1.x * d;
    vout.y = v1.y * d;
    vout.z = v1.z * d;

    pmErrno = 0;
    return 0;
  }


  static public PM_CARTESIAN divide(PM_CARTESIAN v, double d) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartScalDiv(v,d, vout);
    return vout;
  }

  static public int pmCartScalDiv(PmCartesian v1, double d, PmCartesian vout) throws PmException
  {
    if (d == 0.0)
      {

        pmPrintError("Divide by 0 in pmCartScalDiv\n", PM_DIV_ERR);

        vout.x = DBL_MAX;
        vout.y = DBL_MAX;
        vout.z = DBL_MAX;

        pmErrno = PM_DIV_ERR;
        return pmErrno;
      }

    vout.x = v1.x / d;
    vout.y = v1.y / d;
    vout.z = v1.z / d;

    pmErrno = 0;
    return 0;
  }


  static public PM_CARTESIAN neg(PM_CARTESIAN v) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartNeg(v, vout);
    return vout;
  }

  static public int  pmCartNeg(PmCartesian v1, PmCartesian vout) throws PmException
  {
    vout.x = -v1.x;
    vout.y = -v1.y;
    vout.z = -v1.z;

    pmErrno = 0;
    return 0;
  }


  static public PM_CARTESIAN inv(PM_CARTESIAN v) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartInv(v, vout);
    return vout;
  }

  static public int pmCartInv(PmCartesian v1, PmCartesian vout) throws PmException
  {
    double size_sq = pmSq(v1.x) + pmSq(v1.y) + pmSq(v1.z);

    if (size_sq == 0.0)
      {

        pmPrintError("Zero vector in pmCartInv\n", PM_NORM_ERR);


        vout.x = DBL_MAX;
        vout.y = DBL_MAX;
        vout.z = DBL_MAX;

        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    vout.x = v1.x / size_sq;
    vout.y = v1.y / size_sq;
    vout.z = v1.z / size_sq;

    pmErrno = 0;
    return 0;
  }

  static public PM_CARTESIAN norm(PM_CARTESIAN v) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartNorm(v, vout);
    return vout;
  }

  static public int pmCartNorm(PmCartesian v, PmCartesian vout) throws PmException
  {
    double size = pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z));

    if (size == 0.0)
      {

        pmPrintError("Zero vector in pmCartNorm\n", PM_NORM_ERR);


        vout.x = DBL_MAX;
        vout.y = DBL_MAX;
        vout.z = DBL_MAX;

        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    vout.x = v.x / size;
    vout.y = v.y / size;
    vout.z = v.z / size;

    pmErrno = 0;
    return 0;
  }

  static public boolean isNorm(PM_CARTESIAN v) throws PmException
  {
    return pmCartIsNorm(v);
  }

  static public boolean pmCartIsNorm(PmCartesian v) throws PmException
  {
    return pmSqrt(pmSq(v.x) + pmSq(v.y) + pmSq(v.z)) - 1.0 <
      UNIT_VEC_FUZZ;
  }


  static public PM_CARTESIAN proj(PM_CARTESIAN v1, PM_CARTESIAN v2) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmCartCartProj(v1,v2, vout);
    return vout;
  }

  static public int pmCartCartProj(PmCartesian v1, PmCartesian v2, PmCartesian vout) throws PmException
  {
    int r1, r2, r3;
    double d;

    r1  = pmCartNorm(v2,v2);
    d = pmCartCartDot(v1, v2);
    r2 = pmCartScalMult(v2, d, vout);
    r3 = pmErrno;

    if(r1 != 0 || r2 != 0 || r3 != 0)
      {
        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    pmErrno = 0;
    return 0;
  }

   static public int pmCartPlaneProj(PmCartesian v, PmCartesian normal, PmCartesian vout) throws PmException
  {
    int r1, r2;
    PmCartesian par = new PmCartesian();

    r1 = pmCartCartProj(v, normal, par);
    r2 = pmCartCartSub(v, par, vout);

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    pmErrno = 0;
    return 0;
  }

  /* PmCylindrical Functions */


  static public PM_CYLINDRICAL add(PM_CYLINDRICAL c1, PM_CYLINDRICAL c2) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylCylAdd(c1,c2, cout);
    return cout;
  }

  static public int  pmCylCylAdd(PmCylindrical c1, PmCylindrical c2, PmCylindrical cout ) throws PmException
  {
    PmCartesian v1 = new PmCartesian();
    PmCartesian v2 = new PmCartesian();
    PmCartesian vout = new PmCartesian();

    pmCylCartConvert(c1,v1);
    pmCylCartConvert(c2,v2);
    pmCartCartAdd(v1,v2,vout);
    pmCartCylConvert(vout,cout);

     pmErrno = 0;
     return 0;
  }

  static public PM_CYLINDRICAL subtract(PM_CYLINDRICAL c1, PM_CYLINDRICAL c2) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylCylSub(c1,c2,cout);
    return cout;
  }

  static public int pmCylCylSub(PmCylindrical c1, PmCylindrical c2, PmCylindrical cout) throws PmException
  {
    PmCartesian v1 = new PmCartesian();
    PmCartesian v2 = new PmCartesian();
    PmCartesian vout = new PmCartesian();

    pmCylCartConvert(c1,v1);
    pmCylCartConvert(c2,v2);
    pmCartCartSub(v1,v2,vout);
    pmCartCylConvert(vout,cout);

    pmErrno = 0;
    return 0;
  }

  static public PM_CYLINDRICAL multiply(double d, PM_CYLINDRICAL c) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylScalMult(c,d, cout);
    return cout;
  }

  static public PM_CYLINDRICAL multiply(PM_CYLINDRICAL c, double d) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylScalMult(c,d, cout);
    return cout;
  }

  static public int pmCylScalMult(PmCylindrical v1, double d, PmCylindrical vout) throws PmException
  {
    vout.theta = v1.theta;
    vout.r = v1.r * d;
    vout.z = v1.z * d;

    if(vout.r < 0)
      {
        vout.r *= -1;
        if(vout.theta > 0)
          {
            vout.theta -= Math.PI;
          }
        else
          {
            vout.theta += Math.PI;
          }
      }

    pmErrno = 0;
    return 0;
  }

  static public PM_CYLINDRICAL divide(PM_CYLINDRICAL c, double d) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylScalDiv(c,d, cout);
    return cout;
  }


  static public int pmCylScalDiv(PmCylindrical v1, double d, PmCylindrical vout) throws PmException
  {
    if (d == 0.0)
      {

        pmPrintError("Divide by 0 in pmCylScalDiv\n", PM_DIV_ERR);

        vout.r = DBL_MAX;
        vout.z = DBL_MAX;

        pmErrno = PM_DIV_ERR;
        return pmErrno;
      }

    vout.theta = v1.theta;
    vout.r = v1.r / d;
    vout.z = v1.z / d;

     if(vout.r < 0)
      {
        vout.r *= -1;
        if(vout.theta > 0)
          {
            vout.theta -= Math.PI;
          }
        else
          {
            vout.theta += Math.PI;
          }
      }
   pmErrno = 0;
    return 0;
  }

  static public boolean pmCylCylCompare(PmCylindrical v1, PmCylindrical v2) throws PmException
  {
    if(debug_on)
      {
        System.out.println("pmCylCylCompare("+v1+", "+v2+")");
      }
    double angle_diff = Math.abs(v1.theta-v2.theta);
    if(angle_diff  >= V_FUZZ &&
       Math.abs(angle_diff - Math.PI*2) >= V_FUZZ)
      {
           System.out.println("angle_diff = "+angle_diff);
           return false;
      }
    if (
        Math.abs(v1.r-v2.r) >= V_FUZZ ||
        Math.abs(v1.z-v2.z) >= V_FUZZ)
      {
        return false;
      }

    return true;
  }

  static public double dot(PM_CYLINDRICAL v1, PM_CYLINDRICAL v2) throws PmException
  {
    return pmCylCylDot(v1,v2);
  }

  static public double pmCylCylDot(PmCylindrical c1, PmCylindrical c2) throws PmException
  {
    PmCartesian v1 = new PmCartesian();
    PmCartesian v2 = new PmCartesian();

    pmCylCartConvert(c1,v1);
    pmCylCartConvert(c2,v2);
    return pmCartCartDot(v1,v2);
  }

  static public double mag(PM_CYLINDRICAL v1) throws PmException
  {
    return pmCylMag(v1);
  }

  static public double pmCylMag(PmCylindrical c) throws PmException
  {
    return pmSqrt(pmSq(c.r)+pmSq(c.z));
  }

 static public PM_CYLINDRICAL cross(PM_CYLINDRICAL v1, PM_CYLINDRICAL v2) throws PmException
  {
     PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
     pmCylCylCross(v1,v2, cout);
     return cout;
  }


  static public int  pmCylCylCross(PmCylindrical c1, PmCylindrical c2, PmCylindrical cout) throws PmException
  {
    PmCartesian v1 = new PmCartesian();
    PmCartesian v2 = new PmCartesian();
    PmCartesian vout = new PmCartesian();

    pmCylCartConvert(c1,v1);
    pmCylCartConvert(c2,v2);
    pmCartCartCross(v1,v2,vout);
    pmCartCylConvert(vout,cout);

    pmErrno = 0;
    return 0;
  }


  static public PM_CYLINDRICAL neg(PM_CYLINDRICAL v) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylNeg(v, cout);
    return cout;
  }

  static public int  pmCylNeg(PmCylindrical v1, PmCylindrical vout) throws PmException
  {
    vout.theta = v1.theta + Math.PI;
    if(vout.theta > Math.PI)
      {
        vout.theta -= 2*Math.PI;
      }
    vout.r = v1.r;
    vout.z = -v1.z;

    pmErrno = 0;
    return 0;
  }

  static public PM_CYLINDRICAL norm(PM_CYLINDRICAL v) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylNorm(v, cout);
    return cout;
  }

  static public int pmCylNorm(PmCylindrical v, PmCylindrical vout) throws PmException
  {
    double size = pmSqrt( pmSq(v.r) + pmSq(v.z));

    if (size == 0.0)
      {

        pmPrintError("Zero vector in pmCylNorm\n", PM_NORM_ERR);


        vout.r = 0;
        vout.z = 0;

        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    vout.r = v.r / size;
    vout.z = v.z / size;

    pmErrno = 0;
    return 0;
  }

  static public PM_CYLINDRICAL inv(PM_CYLINDRICAL v) throws PmException
  {
    PM_CYLINDRICAL cout = new PM_CYLINDRICAL();
    pmCylInv(v, cout);
    return cout;
  }

  static public int pmCylInv(PmCylindrical c, PmCylindrical cout) throws PmException
  {


    /*
    double size_sq = pmSq(v.r) + pmSq(v.z);

    vout.theta = v.theta;
    if (size_sq == 0.0)
      {

        pmPrintError("Zero vector in pmCylInv\n", PM_NORM_ERR);


        vout.r = DBL_MAX;
        vout.z = DBL_MAX;

        pmErrno = PM_NORM_ERR;
        return pmErrno;
      }

    vout.r = v.r / size_sq;
    vout.z = v.z / size_sq;

    pmErrno = 0;
    */
    PmCartesian v = new PmCartesian();
    pmCylCartConvert(c,v);
    PmCartesian vout = new PmCartesian();
    pmCartInv(v,vout);
    pmCartCylConvert(vout,cout);
    return 0;
  }

  static public boolean isNorm(PM_CYLINDRICAL v) throws PmException
  {
    return pmCylIsNorm(v);
  }

  static public boolean pmCylIsNorm(PmCylindrical v) throws PmException
  {
    return pmSqrt(pmSq(v.r) + pmSq(v.z)) - 1.0 <
      UNIT_VEC_FUZZ;
  }


  /* angle-axis functions */
  static public PmQuaternion pmQuatAxisAngleMult(PmQuaternion q, int axis, double angle) throws PmException
  {
    PmQuaternion pq = new PmQuaternion();
    if(pmQuatAxisAngleMult(q,axis, angle, pq) != 0)
      {
        return null;
      }
    return pq;
  }

  static public int  pmQuatAxisAngleMult(PmQuaternion q, int axis, double angle,  PmQuaternion pq ) throws PmException
  {
    double sh, ch;


    if (! pmQuatIsNorm(q))
      {

        pmPrintError("error: non-unit quaternion in pmQuatAxisAngleMult\n", PM_NORM_ERR);

        return -1;
      }


    angle *= 0.5;
    sh = Math.sin(angle);
    ch = Math.cos(angle);

    switch (axis)
      {
      case PM_X:
        pq.s = ch * q.s - sh * q.x;
        pq.x = ch * q.x + sh * q.s;
        pq.y = ch * q.y + sh * q.z;
        pq.z = ch * q.z - sh * q.y;
        break;

      case PM_Y:
        pq.s = ch * q.s - sh * q.y;
        pq.x = ch * q.x - sh * q.z;
        pq.y = ch * q.y + sh * q.s;
        pq.z = ch * q.z + sh * q.x;
        break;

      case PM_Z:
        pq.s = ch * q.s - sh * q.z;
        pq.x = ch * q.x + sh * q.y;
        pq.y = ch * q.y - sh * q.x;
        pq.z = ch * q.z + sh * q.s;
        break;

      default:

        pmPrintError("error: bad axis in pmQuatAxisAngleMult\n", PM_NORM_ERR);

        return -1;
      }

    if (pq.s < 0.0)
      {
        pq.s *= -1.0;
        pq.x *= -1.0;
        pq.y *= -1.0;
        pq.z *= -1.0;
      }

    return 0;
  }

  /* PmRotationVector functions */

  static public PM_ROTATION_VECTOR  multiply(PM_ROTATION_VECTOR r, double s) throws PmException
  {
    PM_ROTATION_VECTOR rout = new PM_ROTATION_VECTOR();
    pmRotScalMult(r,s, rout);
    return rout;
  }

  static public PmRotationVector  multiply(double s, PM_ROTATION_VECTOR  r) throws PmException
  {
    PM_ROTATION_VECTOR rout = new PM_ROTATION_VECTOR();
    pmRotScalMult(r,s, rout);
    return rout;
  }

  static public int pmRotScalMult(PmRotationVector r, double s, PmRotationVector rout) throws PmException
  {
    rout.s = r.s * s;
    rout.x = r.x;
    rout.y = r.y;
    rout.z = r.z;

    return pmErrno = 0;
  }

  static public PM_ROTATION_VECTOR  divide(PM_ROTATION_VECTOR r, double s) throws PmException
  {
    PM_ROTATION_VECTOR rout = new PM_ROTATION_VECTOR();
    pmRotScalDiv(r,s, rout);
    return rout;
  }

  static public int pmRotScalDiv(PmRotationVector r, double s, PmRotationVector  rout) throws PmException
  {
    if (s == 0.0)
      {

        pmPrintError("Divide by zero in pmRotScalDiv\n", PM_NORM_ERR);


        rout.s = DBL_MAX;
        rout.x = r.x;
        rout.y = r.y;
        rout.z = r.z;

        return pmErrno = PM_NORM_ERR;
      }

    rout.s = r.s / s;
    rout.x = r.x;
    rout.y = r.y;
    rout.z = r.z;

    return pmErrno = 0;
  }

  static public boolean isNorm(PM_ROTATION_VECTOR r) throws PmException
  {
    return pmRotIsNorm(r);
  }

  static public boolean pmRotIsNorm(PmRotationVector r) throws PmException
  {
    if (Math.abs(r.s) < RS_FUZZ ||
        Math.abs(pmSqrt(pmSq(r.x) + pmSq(r.y) + pmSq(r.z))) - 1.0 < UNIT_VEC_FUZZ)
      {
        return true;
      }

    return false;
  }


  static public PM_ROTATION_VECTOR  norm(PM_ROTATION_VECTOR r) throws PmException
  {
    PM_ROTATION_VECTOR rout = new PM_ROTATION_VECTOR();
    pmRotNorm(r, rout);
    return rout;
  }

  static public int pmRotNorm(PmRotationVector r, PmRotationVector rout) throws PmException
  {
    double size;

    size = pmSqrt(pmSq(r.x) + pmSq(r.y) + pmSq(r.z));

    if (Math.abs(r.s) < RS_FUZZ)
      {
        rout.s = 0.0;
        rout.x = 0.0;
        rout.y = 0.0;
        rout.z = 0.0;

        return pmErrno = 0;
      }

    if (size == 0.0)
      {

        pmPrintError("error: pmRotNorm size is zero\n", PM_NORM_ERR);


        rout.s = 0.0;
        rout.x = 0.0;
        rout.y = 0.0;
        rout.z = 0.0;

        return pmErrno = PM_NORM_ERR;
      }

    rout.s = r.s;
    rout.x = r.x / size;
    rout.y = r.y / size;
    rout.z = r.z / size;

    return pmErrno = 0;
  }

  /* PmRotationMatrix functions */
  static public int pmMatNorm(PmRotationMatrix m, PmRotationMatrix mout) throws PmException
  {
    mout = m;
    pmPrintError("error: pmMatNorm not implemented\n", PM_IMPL_ERR);

    return pmErrno = PM_IMPL_ERR;
  }

  static public boolean isNorm(PM_ROTATION_MATRIX m) throws PmException
  {
    return pmMatIsNorm(m);
  }

  static public boolean pmMatIsNorm(PmRotationMatrix m) throws PmException
  {
    PmCartesian u = new PmCartesian();

    pmCartCartCross(m.x, m.y, u);

    return (pmCartIsNorm(m.x) &&
            pmCartIsNorm(m.y) &&
            pmCartIsNorm(m.z) &&
            pmCartCartCompare(u, m.z));
  }

  static public int pmMatInv(PmRotationMatrix m, PmRotationMatrix mout) throws PmException
  {
    /* inverse of a rotation matrix is the transpose */

    mout.x.x = m.x.x;
    mout.x.y = m.y.x;
    mout.x.z = m.z.x;

    mout.y.x = m.x.y;
    mout.y.y = m.y.y;
    mout.y.z = m.z.y;

    mout.z.x = m.x.z;
    mout.z.y = m.y.z;
    mout.z.z = m.z.z;

    return pmErrno = 0;
  }

  static public PM_CARTESIAN  multiply(PM_ROTATION_MATRIX m, PM_CARTESIAN v) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmMatCartMult(m,v, vout);
    return vout;
  }

  static public int pmMatCartMult(PmRotationMatrix m, PmCartesian v, PmCartesian  vout) throws PmException
  {
    vout.x = m.x.x * v.x + m.y.x * v.y + m.z.x * v.z;
    vout.y = m.x.y * v.x + m.y.y * v.y + m.z.y * v.z;
    vout.z = m.x.z * v.x + m.y.z * v.y + m.z.z * v.z;

    return pmErrno = 0;
  }

  static public PM_ROTATION_MATRIX multiply(PM_ROTATION_MATRIX m1, PM_ROTATION_MATRIX m2) throws PmException
  {
    PM_ROTATION_MATRIX mout = new PM_ROTATION_MATRIX();
    pmMatMatMult(m1,m2, mout);
    return mout;
  }

  static public int pmMatMatMult(PmRotationMatrix m1, PmRotationMatrix m2,
                                 PmRotationMatrix  mout) throws PmException
  {
    mout.x.x = m1.x.x * m2.x.x + m1.y.x * m2.x.y + m1.z.x * m2.x.z;
    mout.x.y = m1.x.y * m2.x.x + m1.y.y * m2.x.y + m1.z.y * m2.x.z;
    mout.x.z = m1.x.z * m2.x.x + m1.y.z * m2.x.y + m1.z.z * m2.x.z;

    mout.y.x = m1.x.x * m2.y.x + m1.y.x * m2.y.y + m1.z.x * m2.y.z;
    mout.y.y = m1.x.y * m2.y.x + m1.y.y * m2.y.y + m1.z.y * m2.y.z;
    mout.y.z = m1.x.z * m2.y.x + m1.y.z * m2.y.y + m1.z.z * m2.y.z;

    mout.z.x = m1.x.x * m2.z.x + m1.y.x * m2.z.y + m1.z.x * m2.z.z;
    mout.z.y = m1.x.y * m2.z.x + m1.y.y * m2.z.y + m1.z.y * m2.z.z;
    mout.z.z = m1.x.z * m2.z.x + m1.y.z * m2.z.y + m1.z.z * m2.z.z;

    return pmErrno = 0;
  }

  /* PmQuaternion functions */

  static public boolean pmQuatQuatCompare(PmQuaternion q1, PmQuaternion q2) throws PmException
  {

    if (! pmQuatIsNorm(q1) ||
        ! pmQuatIsNorm(q2))
      {

        pmPrintError("Bad quaternion in pmQuatQuatCompare\n", PM_ERR);

      }


    if (Math.abs(q1.s-q2.s) < Q_FUZZ &&
        Math.abs(q1.x-q2.x) < Q_FUZZ &&
        Math.abs(q1.y-q2.y) < Q_FUZZ &&
        Math.abs(q1.z-q2.z) < Q_FUZZ)
      {
        return true;
      }

    /* note (0, x, y, z) = (0, -x, -y, -z) */
    if (Math.abs(q1.s) >= QS_FUZZ ||
        Math.abs(q1.x + q2.x) >= Q_FUZZ ||
        Math.abs(q1.y + q2.y) >= Q_FUZZ ||
        Math.abs(q1.z + q2.z) >= Q_FUZZ)
      {
        return false;
      }

    return true;
  }

  static public double mag(PM_QUATERNION q) throws PmException
  {
    return pmQuatMag(q);
  }

  static public double pmQuatMag(PmQuaternion q) throws PmException
  {
    double d;
    PmRotationVector r = new PmRotationVector();

    pmQuatRotConvert(q, r);
    d = r.s;
    return d;

  }

  static public PM_QUATERNION norm(PM_QUATERNION q1 ) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatNorm(q1, qout);
    return qout;
  }

  static public int pmQuatNorm(PmQuaternion q1, PmQuaternion qout) throws PmException
  {
    double size = pmSqrt(pmSq(q1.s) + pmSq(q1.x) + pmSq(q1.y) + pmSq(q1.z));

    if (size == 0.0)
      {

        pmPrintError("Bad quaternion in pmQuatNorm\n", PM_NORM_ERR);

        qout.s = 1;
        qout.x = 0;
        qout.y = 0;
        qout.z = 0;

        return pmErrno = PM_NORM_ERR;
      }

    if (q1.s >= 0.0)
      {
        qout.s = q1.s / size;
        qout.x = q1.x / size;
        qout.y = q1.y / size;
        qout.z = q1.z / size;

        return pmErrno = 0;
      }
    else
      {
        qout.s = -q1.s / size;
        qout.x = -q1.x / size;
        qout.y = -q1.y / size;
        qout.z = -q1.z / size;

        return pmErrno = 0;
      }
  }

  static public PM_QUATERNION inv(PM_QUATERNION q1 ) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatInv(q1, qout);
    return qout;
  }

  static public int pmQuatInv(PmQuaternion q1, PmQuaternion qout) throws PmException
  {
    qout.s = q1.s;
    qout.x = - q1.x;
    qout.y = - q1.y;
    qout.z = - q1.z;


    if (! pmQuatIsNorm(q1))
      {

        pmPrintError("Bad quaternion in pmQuatInv\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    return pmErrno = 0;
  }

  static public boolean isNorm(PM_QUATERNION q1) throws PmException
  {
    return pmQuatIsNorm(q1);
  }

  static public boolean pmQuatIsNorm(PmQuaternion q1) throws PmException
  {
    return (Math.abs(pmSq(q1.s) + pmSq(q1.x) + pmSq(q1.y) + pmSq(q1.z) - 1.0) <
            UNIT_QUAT_FUZZ);
  }

  static public PM_QUATERNION multiply(PM_QUATERNION q, double s) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatScalMult(q,s, qout);
    return qout;
  }

  static public PM_QUATERNION multiply(double s, PM_QUATERNION q) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatScalMult(q,s, qout);
    return qout;
  }

  static public int pmQuatScalMult(PmQuaternion q, double s, PmQuaternion qout) throws PmException
  {
    /* FIXME-- need a native version; this goes through a rotation vector */
    PmRotationVector r = new PmRotationVector();
    int r1, r2, r3;

    r1 = pmQuatRotConvert(q, r);
    r2 = pmRotScalMult(r, s, r);
    r3 = pmRotQuatConvert(r, qout);

    if(r1 != 0 || r2 != 0 || r3 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public PM_QUATERNION divide(PM_QUATERNION q, double s) throws PmException
  {
     PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatScalDiv(q,s, qout);
    return qout;
  }

  static public int pmQuatScalDiv(PmQuaternion q, double s, PmQuaternion qout) throws PmException
  {
    /* FIXME-- need a native version; this goes through a rotation vector */
    PmRotationVector r = new PmRotationVector();
    int r1, r2, r3;

    r1 = pmQuatRotConvert(q, r);
    r2 = pmRotScalDiv(r, s, r);
    r3 = pmRotQuatConvert(r, qout);

    if(r1 != 0 || r2 != 0 || r3 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public PM_QUATERNION multiply(PM_QUATERNION q1, PM_QUATERNION q2 ) throws PmException
  {
    PM_QUATERNION qout = new PM_QUATERNION();
    pmQuatQuatMult(q1,q2, qout);
    return qout;
  }

  static public int pmQuatQuatMult(PmQuaternion q1, PmQuaternion q2, PmQuaternion qout) throws PmException
  {
    qout.s = q1.s * q2.s - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z;

    if (qout.s >= 0.0)
      {
        qout.x = q1.s * q2.x + q1.x * q2.s + q1.y * q2.z - q1.z * q2.y;
        qout.y = q1.s * q2.y - q1.x * q2.z + q1.y * q2.s + q1.z * q2.x;
        qout.z = q1.s * q2.z + q1.x * q2.y - q1.y * q2.x + q1.z * q2.s;
      }
    else
      {
        qout.s *= -1;
        qout.x = - q1.s * q2.x - q1.x * q2.s - q1.y * q2.z + q1.z * q2.y;
        qout.y = - q1.s * q2.y + q1.x * q2.z - q1.y * q2.s - q1.z * q2.x;
        qout.z = - q1.s * q2.z - q1.x * q2.y + q1.y * q2.x - q1.z * q2.s;
      }


    if (! pmQuatIsNorm(q1) ||
        ! pmQuatIsNorm(q2))
      {

        pmPrintError("Bad quaternion in pmQuatQuatMult\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    return pmErrno = 0;
  }

  static public PM_CARTESIAN multiply(PM_QUATERNION q1, PM_CARTESIAN v2 ) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmQuatCartMult(q1,v2,vout);
    return vout;
  }

  static public int pmQuatCartMult(PmQuaternion q1, PmCartesian v2, PmCartesian vout) throws PmException
  {
    PmCartesian c = new PmCartesian();

    c.x = q1.y * v2.z  -  q1.z * v2.y;
    c.y = q1.z * v2.x  -  q1.x * v2.z;
    c.z = q1.x * v2.y  -  q1.y * v2.x;

    vout.x = v2.x + 2.0 * (q1.s * c.x + q1.y * c.z - q1.z * c.y);
    vout.y = v2.y + 2.0 * (q1.s * c.y + q1.z * c.x - q1.x * c.z);
    vout.z = v2.z + 2.0 * (q1.s * c.z + q1.x * c.y - q1.y * c.x);


    if (! pmQuatIsNorm(q1))
      {

        pmPrintError("Bad quaternion in pmQuatCartMult\n",PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    return pmErrno = 0;
  }

  /* PmPose functions*/

  static public boolean pmPosePoseCompare(PmPose p1, PmPose p2) throws PmException
  {

    if (! pmQuatIsNorm(p1.rot) ||
        ! pmQuatIsNorm(p2.rot))
      {

        pmPrintError("Bad quaternion in pmPosePoseCompare\n",PM_ERR);

      }


    return( pmQuatQuatCompare(p1.rot, p2.rot) &&
                      pmCartCartCompare(p1.tran, p2.tran));
  }

  static public PM_POSE inv(PM_POSE p) throws PmException
  {
    PM_POSE pout = new PM_POSE();
    pmPoseInv(p,pout);
    return pout;
  }

  static public int pmPoseInv(PmPose p1, PmPose p2) throws PmException
  {
    int r1, r2;


    if (! pmQuatIsNorm(p1.rot))
      {

        pmPrintError("Bad quaternion in pmPoseInv\n", PM_ERR);

      }


    r1 = pmQuatInv(p1.rot, p2.rot);
    r2 = pmQuatCartMult(p2.rot, p1.tran, p2.tran);

    p2.tran.x *= -1.0;
    p2.tran.y *= -1.0;
    p2.tran.z *= -1.0;

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public PM_CARTESIAN multiply(PM_POSE p1, PM_CARTESIAN v2) throws PmException
  {
    PM_CARTESIAN vout = new PM_CARTESIAN();
    pmPoseCartMult(p1,v2,vout);
    return vout;
  }

  static public int pmPoseCartMult(PmPose p1, PmCartesian v2, PmCartesian vout) throws PmException
  {
    int r1, r2;


    if (! pmQuatIsNorm(p1.rot))
      {

        pmPrintError("Bad quaternion in pmPoseCartMult\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    r1 = pmQuatCartMult(p1.rot, v2, vout);
    r2 = pmCartCartAdd(p1.tran, vout, vout);

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public PM_POSE multiply(PM_POSE p1, PM_POSE p2) throws PmException
  {
    PM_POSE pout = new PM_POSE();
    pmPosePoseMult(p1,p2,pout);
    return pout;
  }

 static public int pmPosePoseMult(PmPose p1, PmPose p2, PmPose pout) throws PmException
  {
    int r1, r2, r3;


    if (! pmQuatIsNorm(p1.rot) ||
        ! pmQuatIsNorm(p2.rot))
      {

        pmPrintError("Bad quaternion in pmPosePoseMult\n", PM_NORM_ERR);

        return pmErrno = PM_NORM_ERR;
      }


    r1 = pmQuatCartMult(p1.rot, p2.tran, pout.tran);
    r2 = pmCartCartAdd(p1.tran, pout.tran, pout.tran);
    r3 = pmQuatQuatMult(p1.rot, p2.rot, pout.rot);

    if(r1 != 0 || r2 != 0 || r3 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  /* homogeneous transform functions */
  static public PM_HOMOGENEOUS inv(PM_HOMOGENEOUS h) throws PmException
  {
    PM_HOMOGENEOUS hout = new PM_HOMOGENEOUS();
    pmHomInv(h,hout);
    return hout;
  }

  static public int pmHomInv(PmHomogeneous h1, PmHomogeneous h2) throws PmException
  {
    int r1, r2;


    if (! pmMatIsNorm(h1.rot))
      {

        pmPrintError("Bad rotation matrix in pmHomInv\n", PM_ERR);

      }


    r1 = pmMatInv(h1.rot, h2.rot);
    r2 = pmMatCartMult(h2.rot, h1.tran, h2.tran);

    h2.tran.x *= -1.0;
    h2.tran.y *= -1.0;
    h2.tran.z *= -1.0;

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;

  }

  /* line functions */

  static public int pmLineInit(PmLine line, PmPose start, PmPose end) throws PmException
  {
    int r1, r2 = 0;
    double mag;

    line.start = start;
    line.end = end;
    r1 = pmCartCartSub(end.tran, start.tran, line.uVec);
    mag = pmCartMag(line.uVec);
    if (IS_FUZZ(mag, CART_FUZZ))
      {
        line.uVec.x = 1.0;
        line.uVec.y = 0.0;
        line.uVec.z = 0.0;
      }
    else
      {
        r2 = pmCartNorm(line.uVec, line.uVec);
      }

    /* return PM_NORM_ERR if uVec has been set to 1, 0, 0 */
    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  static public int pmLinePoint(PmLine line, double len, PmPose point) throws PmException
  {
    int r1, r2;

    /* return start + len * uVec */
    r1 = pmCartScalMult(line.uVec, len, point.tran);
    r2 = pmCartCartAdd(line.start.tran, point.tran, point.tran);

    if(r1 != 0 || r2 != 0)
      {
        pmErrno = PM_NORM_ERR;
      }
    else
      {
        pmErrno = 0;
      }

    return pmErrno;
  }

  /* circle functions */

  /*
    pmCircleInit() takes the defining parameters of a generalized circle
    and sticks them in the structure. It also computes the radius and vectors
    in the plane that are useful for other functions and that don't need
    to be recomputed every time.

    Note that the end can be placed arbitrarily, resulting in a combination of
    spiral and helical motion. There is an overconstraint between the start,
    center, and normal vector: the center vector and start vector are assumed
    to be in the plane defined by the normal vector. If this is not true, then
    it will be made true by moving the center vector onto the plane.
    */
  static public int pmCircleInit(PmCircle circle,
                                 PmPose start, PmPose end,
                                 PmCartesian center, PmCartesian normal,
                                 int turn) throws PmException
  {
    double dot;
    PmCartesian rEnd = new PmCartesian();
    PmCartesian v = new PmCartesian();
    double d;
    int r1;


    if (null == circle)
      {

        pmPrintError("error: pmCircleInit cirle pointer is null\n", PM_ERR);

        return pmErrno = PM_ERR;
      }


    /* adjust center */
    pmCartCartSub(start.tran, center, v);
    r1 = pmCartCartProj(v, normal, v);
    if (PM_NORM_ERR == r1)
      {
        /* bad normal vector-- abort */

        pmPrintError("error: pmCircleInit normal vector is 0\n", PM_ERR);

        return -1;
      }
    pmCartCartAdd(v, center, circle.center);

    /* normalize and redirect normal vector based on turns. If turn is
       less than 0, point normal vector in other direction and make
       turn positive, -1 . 0, -2 . 1, etc. */
    pmCartNorm(normal, circle.normal);
    if (turn < 0)
      {
        turn = -1 - turn;
        pmCartScalMult(circle.normal, -1.0, circle.normal);
      }

    /* radius */
   circle.radius = pmCartCartDisp(start.tran, circle.center);

    /* vector in plane of circle from center to start, magnitude radius */
    pmCartCartSub(start.tran, circle.center, circle.rTan);
    /* vector in plane of circle perpendicular to rTan, magnitude radius */
    pmCartCartCross(circle.normal, circle.rTan, circle.rPerp);

    /* do rHelix, rEnd */
    pmCartCartSub(end.tran, circle.center, circle.rHelix);
    pmCartPlaneProj(circle.rHelix, circle.normal, rEnd);
    circle.spiral = pmCartMag(rEnd);
    circle.spiral -= circle.radius;
    pmCartCartSub(circle.rHelix, rEnd, circle.rHelix);
    pmCartNorm(rEnd, rEnd);
    pmCartScalMult(rEnd, circle.radius, rEnd);

    /* angle */
    dot = pmCartCartDot(circle.rTan, rEnd );
    dot = dot / (circle.radius * circle.radius);
    circle.angle = Math.acos(dot);
    /* now angle is in range 0..PI . Check if cross is antiparallel to
       normal. If so, true angle is between PI..2PI. Need to subtract
       from 2PI. */
    pmCartCartCross(circle.rTan, rEnd, v);
    d = pmCartCartDot(v, circle.normal);
    if (d < 0.0)
      {
        circle.angle = PM_2_PI - circle.angle;
      }

    /* now add more angle for multi turns */
    if (turn > 0)
      {
        circle.angle += turn * 2.0 * PM_PI;
      }

    return pmErrno = 0;
  }

  /*
    pmCirclePoint() returns the vector to the point at the given angle along
    the circle. If the circle is a helix or spiral or combination, the
    point will include interpolation off the actual circle.
    */
  static public int pmCirclePoint(PmCircle circle, double angle, PmPose point) throws PmException
  {
    PmCartesian par = new PmCartesian();
    PmCartesian perp = new PmCartesian();
    double scale;


    if (null == circle ||
        null == point)
      {

        pmPrintError("error: pmCirclePoint circle or point pointer is null\n", PM_ERR);

        return pmErrno = PM_ERR;
      }


    /* compute components rel to center */
    pmCartScalMult(circle.rTan, Math.cos(angle), par);
    pmCartScalMult(circle.rPerp, Math.sin(angle), perp);

    /* add to get radius vector rel to center */
    pmCartCartAdd(par, perp, point.tran);

    /* get scale for spiral, helix interpolation */

    if (circle.angle == 0.0)
      {
        pmPrintError("error: pmCirclePoint angle is zero\n", PM_ERR);

        return pmErrno = PM_DIV_ERR;
      }

    scale = angle / circle.angle;

    /* add scaled vector in radial dir for spiral */
    pmCartNorm(point.tran, par);
    pmCartScalMult(par, scale * circle.spiral, par);
    pmCartCartAdd(point.tran, par, point.tran);

    /* add scaled vector in helix dir */
    pmCartScalMult(circle.rHelix, scale, perp);
    pmCartCartAdd(point.tran, perp, point.tran);

    /* add to center vector for final result */
    pmCartCartAdd(circle.center, point.tran, point.tran);

    return pmErrno = 0;
  }
}
